from .constants import (
    NORMAL, FONT_STYLES, FONT_VARIANTS, FONT_WEIGHTS,
    ITALIC, OBLIQUE, SMALL_CAPS, BOLD
)

class Font:
    def __init__(self, family, size, style=NORMAL, variant=NORMAL, weight=NORMAL):
        if (family[0] == "'" and family[-1] == "'") or (family[0] == '"' and family[-1] == '"'):
            self.family = family[1:-1]
        else:
            self.family = family

        try:
            self.size = int(size)
        except ValueError:
            try:
                if size.strip().endswith('pt'):
                    self.size = int(size[:-2])
                else:
                    raise ValueError("Invalid font size {!r}".format(size))
            except:
                raise ValueError("Invalid font size {!r}".format(size))
        self.style = style if style in FONT_STYLES else NORMAL
        self.variant = variant if variant in FONT_VARIANTS else NORMAL
        self.weight = weight if weight in FONT_WEIGHTS else NORMAL

    def __repr__(self):
        return '<Font: {}{}{}{}pt {}>'.format(
            '' if self.style is NORMAL else (self.style + ' '),
            '' if self.variant is NORMAL else (self.variant + ' '),
            '' if self.weight is NORMAL else (self.weight + ' '),
            self.size,
            self.family
        )

    def __eq__(self, other):
        return (
            self.family == other.family
            and self.size == other.size
            and self.style == other.style
            and self.variant == other.variant
            and self.weight == other.weight
        )

    def normal_style(self):
        "Generate a normal style version of this font"
        return Font(self.family, self.size, style=NORMAL, variant=self.variant, weight=self.weight)

    def italic(self):
        "Generate an italic version of this font"
        return Font(self.family, self.size, style=ITALIC, variant=self.variant, weight=self.weight)

    def oblique(self):
        "Generate an oblique version of this font"
        return Font(self.family, self.size, style=OBLIQUE, variant=self.variant, weight=self.weight)

    def normal_variant(self):
        "Generate a normal variant of this font"
        return Font(self.family, self.size, style=self.style, variant=NORMAL, weight=self.weight)

    def small_caps(self):
        "Generate a small-caps variant of this font"
        return Font(self.family, self.size, style=self.style, variant=SMALL_CAPS, weight=self.weight)

    def normal_weight(self):
        "Generate a normal weight version of this font"
        return Font(self.family, self.size, style=self.style, variant=self.variant, weight=NORMAL)

    def bold(self):
        "Generate a bold version of this font"
        return Font(self.family, self.size, style=self.style, variant=self.variant, weight=BOLD)


def font(value):
    """Parse a font from a string.

    Accepts:
    * Font instances

    style: normal / italic / oblique
    variant: normal / small-caps
    weight: normal / bold

    style variant weight size family
    variant weight size family
    weight size family
    size family
    """

    if isinstance(value, Font):
        return value

    elif isinstance(value, str):
        parts = value.split(' ')

        style = None
        variant = None
        weight = None
        size = None

        while size is None:
            part = parts.pop(0)
            if part == NORMAL:
                if style is None:
                    style = NORMAL
                elif variant is None:
                    variant = NORMAL
                elif weight is None:
                    weight = NORMAL
            elif part in FONT_STYLES:
                if style is not None:
                    raise ValueError("Invalid font declaration '{}'".format(value))
                style = part
            elif part in FONT_VARIANTS:
                if variant is not None:
                    raise ValueError("Invalid font declaration '{}'".format(value))
                if style is None:
                     style = NORMAL
                variant = part
            elif part in FONT_WEIGHTS:
                if weight is not None:
                    raise ValueError("Invalid font declaration '{}'".format(value))
                if style is None:
                     style = NORMAL
                if variant is None:
                     variant = NORMAL
                weight = part
            else:
                try:
                    if part.endswith('pt'):
                        size = int(part[:-2])
                    else:
                        size = int(part)
                except ValueError:
                    raise ValueError("Invalid size in font declaration '{}'".format(value))

                if parts[0] == 'pt':
                    parts.pop(0)

        family = ' '.join(parts)
        return Font(family, size, style=style, variant=variant, weight=weight)

    raise ValueError("Unknown font '%s'" % value)



from unittest import TestCase

from travertino.constants import NORMAL, ITALIC, OBLIQUE, SMALL_CAPS, BOLD
from travertino.fonts import font, Font


class FontTests(TestCase):
    def assertFont(self, font, family, size, style, variant, weight):
        self.assertEqual(font.family, family)
        self.assertEqual(font.size, size)
        self.assertEqual(font.style, style)
        self.assertEqual(font.variant, variant)
        self.assertEqual(font.weight, weight)

    def test_equality(self):
        self.assertEqual(
            Font('Comic Sans', '12 pt'),
            Font('Comic Sans', 12, NORMAL, NORMAL, NORMAL)
        )

    def test_repr(self):
        self.assertEqual(
            repr(Font('Comic Sans', 12)),
            '<Font: 12pt Comic Sans>'
        )

        self.assertEqual(
            repr(Font('Comic Sans', 12, style=ITALIC)),
            '<Font: italic 12pt Comic Sans>'
        )

        self.assertEqual(
            repr(Font('Comic Sans', 12, style=ITALIC, variant=SMALL_CAPS)),
            '<Font: italic small-caps 12pt Comic Sans>'
        )

        self.assertEqual(
            repr(Font('Comic Sans', 12, style=ITALIC, variant=SMALL_CAPS, weight=BOLD)),
            '<Font: italic small-caps bold 12pt Comic Sans>'
        )

        self.assertEqual(
            repr(Font('Comic Sans', 12, variant=SMALL_CAPS, weight=BOLD)),
            '<Font: small-caps bold 12pt Comic Sans>'
        )

        self.assertEqual(
            repr(Font('Comic Sans', 12, weight=BOLD)),
            '<Font: bold 12pt Comic Sans>'
        )

        self.assertEqual(
            repr(Font('Comic Sans', 12, style=ITALIC, weight=BOLD)),
            '<Font: italic bold 12pt Comic Sans>'
        )

    def test_simple_construction(self):
        # Simplest case
        self.assertFont(
            Font('Comic Sans', 12),
            'Comic Sans', 12, NORMAL, NORMAL, NORMAL
        )

        # String size
        self.assertFont(
            Font('Comic Sans', '12'),
            'Comic Sans', 12, NORMAL, NORMAL, NORMAL
        )

        # String size with 'pt'
        self.assertFont(
            Font('Comic Sans', '12pt'),
            'Comic Sans', 12, NORMAL, NORMAL, NORMAL
        )

        self.assertFont(
            Font('Comic Sans', '12 pt'),
            'Comic Sans', 12, NORMAL, NORMAL, NORMAL
        )

        with self.assertRaises(ValueError):
            Font('Comic Sans', '12 quatloos'),

    def test_family(self):
        self.assertFont(
            Font('Comic Sans', 12),
            'Comic Sans', 12, NORMAL, NORMAL, NORMAL
        )

        self.assertFont(
            Font('Wingdings', 12),
            'Wingdings', 12, NORMAL, NORMAL, NORMAL
        )

        self.assertFont(
            Font("'Comic Sans'", 12),
            'Comic Sans', 12, NORMAL, NORMAL, NORMAL
        )

        self.assertFont(
            Font('"Comic Sans"', 12),
            'Comic Sans', 12, NORMAL, NORMAL, NORMAL
        )

    def test_style(self):
        self.assertFont(
            Font('Comic Sans', 12, style=ITALIC),
            'Comic Sans', 12, ITALIC, NORMAL, NORMAL
        )

        self.assertFont(
            Font('Comic Sans', 12, style='italic'),
            'Comic Sans', 12, ITALIC, NORMAL, NORMAL
        )

        self.assertFont(
            Font('Comic Sans', 12, style=OBLIQUE),
            'Comic Sans', 12, OBLIQUE, NORMAL, NORMAL
        )

        self.assertFont(
            Font('Comic Sans', 12, style='oblique'),
            'Comic Sans', 12, OBLIQUE, NORMAL, NORMAL
        )

        self.assertFont(
            Font('Comic Sans', 12, style='something else'),
            'Comic Sans', 12, NORMAL, NORMAL, NORMAL
        )

    def test_make_normal_style(self):
        f = Font('Comic Sans', 12)
        self.assertFont(
            f.normal_style(),
            'Comic Sans', 12, NORMAL, NORMAL, NORMAL
        )

        f = Font('Comic Sans', 12, style=ITALIC)
        self.assertFont(
            f.normal_style(),
            'Comic Sans', 12, NORMAL, NORMAL, NORMAL
        )

    def test_make_italic(self):
        f = Font('Comic Sans', 12)
        self.assertFont(
            f.italic(),
            'Comic Sans', 12, ITALIC, NORMAL, NORMAL
        )

    def test_make_oblique(self):
        f = Font('Comic Sans', 12)
        self.assertFont(
            f.oblique(),
            'Comic Sans', 12, OBLIQUE, NORMAL, NORMAL
        )

    def test_variant(self):
        self.assertFont(
            Font('Comic Sans', 12, variant=SMALL_CAPS),
            'Comic Sans', 12, NORMAL, SMALL_CAPS, NORMAL
        )

        self.assertFont(
            Font('Comic Sans', 12, variant='small-caps'),
            'Comic Sans', 12, NORMAL, SMALL_CAPS, NORMAL
        )

        self.assertFont(
            Font('Comic Sans', 12, variant='something else'),
            'Comic Sans', 12, NORMAL, NORMAL, NORMAL
        )

    def test_make_normal_variant(self):
        f = Font('Comic Sans', 12)
        self.assertFont(
            f.normal_variant(),
            'Comic Sans', 12, NORMAL, NORMAL, NORMAL
        )

        f = Font('Comic Sans', 12, variant=SMALL_CAPS)
        self.assertFont(
            f.normal_variant(),
            'Comic Sans', 12, NORMAL, NORMAL, NORMAL
        )

    def test_make_small_caps(self):
        f = Font('Comic Sans', 12)
        self.assertFont(
            f.small_caps(),
            'Comic Sans', 12, NORMAL, SMALL_CAPS, NORMAL
        )

    def test_weight(self):
        self.assertFont(
            Font('Comic Sans', 12, weight=BOLD),
            'Comic Sans', 12, NORMAL, NORMAL, BOLD
        )

        self.assertFont(
            Font('Comic Sans', 12, weight='bold'),
            'Comic Sans', 12, NORMAL, NORMAL, BOLD
        )

        self.assertFont(
            Font('Comic Sans', 12, weight='something else'),
            'Comic Sans', 12, NORMAL, NORMAL, NORMAL
        )

    def test_make_normal_weight(self):
        f = Font('Comic Sans', 12)
        self.assertFont(
            f.normal_weight(),
            'Comic Sans', 12, NORMAL, NORMAL, NORMAL
        )

        f = Font('Comic Sans', 12, weight=BOLD)
        self.assertFont(
            f.normal_weight(),
            'Comic Sans', 12, NORMAL, NORMAL, NORMAL
        )

    def test_make_bold(self):
        f = Font('Comic Sans', 12)
        self.assertFont(
            f.bold(),
            'Comic Sans', 12, NORMAL, NORMAL, BOLD
        )


class ParseFontTests(TestCase):
    def assertFont(self, font, family, size, style, variant, weight):
        self.assertEqual(font.family, family)
        self.assertEqual(font.size, size)
        self.assertEqual(font.style, style)
        self.assertEqual(font.variant, variant)
        self.assertEqual(font.weight, weight)

    def test_font_instance(self):
        f = Font('Comic Sans', 12)

        parsed = font(f)

        self.assertEqual(f, parsed)
        self.assertIs(f, parsed)

    def test_successful_combinations(self):
        self.assertFont(
            font('12pt Comic Sans'),
            'Comic Sans', 12, NORMAL, NORMAL, NORMAL
        )

        self.assertFont(
            font('italic 12pt Comic Sans'),
            'Comic Sans', 12, ITALIC, NORMAL, NORMAL
        )

        self.assertFont(
            font('italic small-caps 12pt Comic Sans'),
            'Comic Sans', 12, ITALIC, SMALL_CAPS, NORMAL
        )

        self.assertFont(
            font('italic small-caps bold 12pt Comic Sans'),
            'Comic Sans', 12, ITALIC, SMALL_CAPS, BOLD
        )

        self.assertFont(
            font('small-caps bold 12pt Comic Sans'),
            'Comic Sans', 12, NORMAL, SMALL_CAPS, BOLD
        )

        self.assertFont(
            font('italic bold 12 pt Comic Sans'),
            'Comic Sans', 12, ITALIC, NORMAL, BOLD
        )

        self.assertFont(
            font('bold 12 pt Comic Sans'),
            'Comic Sans', 12, NORMAL, NORMAL, BOLD
        )

    def test_font_sizes(self):
        self.assertFont(
            font('12pt Comic Sans'),
            'Comic Sans', 12, NORMAL, NORMAL, NORMAL
        )

        self.assertFont(
            font('12 pt Comic Sans'),
            'Comic Sans', 12, NORMAL, NORMAL, NORMAL
        )

        self.assertFont(
            font('12 Comic Sans'),
            'Comic Sans', 12, NORMAL, NORMAL, NORMAL
        )

        with self.assertRaises(ValueError):
            font('12quatloo Comic Sans')

    def test_font_family(self):
        self.assertFont(
            font("12pt 'Comic Sans'"),
            'Comic Sans', 12, NORMAL, NORMAL, NORMAL
        )

        self.assertFont(
            font('12pt "Comic Sans"'),
            'Comic Sans', 12, NORMAL, NORMAL, NORMAL
        )

    def test_normal(self):
        self.assertFont(
            font('normal 12pt Comic Sans'),
            'Comic Sans', 12, NORMAL, NORMAL, NORMAL
        )

        self.assertFont(
            font('italic normal 12pt Comic Sans'),
            'Comic Sans', 12, ITALIC, NORMAL, NORMAL
        )

        self.assertFont(
            font('italic small-caps normal 12pt Comic Sans'),
            'Comic Sans', 12, ITALIC, SMALL_CAPS, NORMAL
        )

    def test_style(self):
        self.assertFont(
            font('italic 12pt Comic Sans'),
            'Comic Sans', 12, ITALIC, NORMAL, NORMAL
        )

        self.assertFont(
            font('oblique 12pt Comic Sans'),
            'Comic Sans', 12, OBLIQUE, NORMAL, NORMAL
        )

        with self.assertRaises(ValueError):
            font('wiggly small-caps bold 12pt Comic Sans')

    def test_variant(self):
        self.assertFont(
            font('italic small-caps 12pt Comic Sans'),
            'Comic Sans', 12, ITALIC, SMALL_CAPS, NORMAL
        )

        with self.assertRaises(ValueError):
            font('italic wiggly bold 12pt Comic Sans')

    def test_weight(self):
        self.assertFont(
            font('italic small-caps bold 12pt Comic Sans'),
            'Comic Sans', 12, ITALIC, SMALL_CAPS, BOLD
        )

        with self.assertRaises(ValueError):
            font('italic small-caps wiggly 12pt Comic Sans')

    def test_duplicates(self):
        with self.assertRaises(ValueError):
            font('oblique italic 12pt Comic Sans')

        with self.assertRaises(ValueError):
            font('italic small-caps oblique 12pt Comic Sans')

        with self.assertRaises(ValueError):
            font('italic small-caps bold small-caps 12pt Comic Sans')

        with self.assertRaises(ValueError):
            font('bold bold 12pt Comic Sans')

    def test_invaid(self):
        with self.assertRaises(ValueError):
            font(42)
